import Vuex from 'vuex'

class VuexStorage {

  /**
   * 构造器
   * @method generateVuexInstance
   * @return {Object} Vuex Modules 对象
   */
  constructor(whiteList = []) {
    /**
     * @member {Array} whiteList clear 的白名单
     */
    this.whiteList = []
    /**
     * @member {Object} $store vuex store
     */
    this.$store = null
    /**
     * @member {boolean} 是否获取备用仓库
     */
    this.getBackUp = false
    /**
     * @member {string} UPDATE_ITEM 更新方法 key 值
     */
    this.UPDATE_ITEM = 'DEFAULT'
    /**
     * @member {string} UPDATE_ITEM 更新方法 key 值
     */
    this.innerOperation = false
  }

  /**
   * 生成静态模块
   * @method generateVuexInstance
   * @return {Object} Vuex Modules 对象
   */
  generateVuexInstance() {
    return {
      getters: {
        getItem: () => (key, isTemp) => {
          const { item } = this.checkItem(key, isTemp)
          return item
        }
      },
      mutations: {
        [this.UPDATE_ITEM]: (state, {
          key,
          value,
          isTemp,
        }) => {
          // 如果为字符串，需要做转化为Array的特殊处理
          if (typeof key === 'string') {
            key = [key]
          }

          let lastKey = key.pop()

          this.checkItem(key, isTemp, false, {
            lastKey,
            value
          })
        },
      },
      modules: {},
    }
  }

  /**
   * 存入store
   * @method setItem
   * @param {string|Array} key 需要存储的键
   * @param {any} value 需要存储的值
   * @param {boolean} [isTemp=true] 是否存储到临时的仓库，默认为 true
   */
  setItem(key, value, isTemp = false) {
    if (!this.innerOperation) {
      const keys = [].concat(key)
      keys.pop()
      this.checkItem(keys, isTemp, true)
    }
    VuexStorage.loopCallBack((store, index) => {
      const setKey = store.concat(key)
      let setValue = value || {}
      if (store.indexOf('backup') > -1) {
        this.getBackUp = true
        const backupValue = this.getItem(key)
        this.getBackUp = false
        if (backupValue) {
          setValue = backupValue
        }
      }

      this.$store.registerModule(setKey, { state: setValue })
    }, isTemp)

    this.innerOperation = false
  }

  /**
   * 实例方法，用于快速增加 item 列表和更新白名单
   * @method hyperChannel
   * @param {Array} [whiteList=[]] 路由白名单
   * @param {boolean} [isTemp=true] 是否存储到临时的仓库，默认为 true
   * @param {RestObejct} items rest 风格 item 列表
   *
   * @memberOf VuexStorage
   */
  // hyperChannel(whiteList = [], isTemp = true, ...items) {
  //   this.whiteList = whiteList
  //   // for (let { key, value } of items) {
  //   //   const temp = this.getItem(key, isTemp)
  //   //   if (temp) {
  //   //     if (value instanceof Object) {
  //   //       Object.assign(value, temp)
  //   //     } else if (value instanceof Array) {
  //   //       value.splice(0, value.length, ...temp)
  //   //     } else {
  //   //       value = temp
  //   //     }
  //   //   }
  //   this.setItem(key, value, isTemp)
  //   // }
  // }

  /**
   * 获取仓库中的指定值
   * @method getItem
   * @param {string|Array} key 仓库中的键
   * @param {boolean} [isTemp=true] 是否存储到临时的仓库，默认为 true
   * @return {any} 仓库中存储的值
   *
   * @memberOf VuexStorage
   */
  getItem(key, isTemp = true) {
    return this.$store.getters.getItem(key, isTemp)
  }

  /**
   * 删除指定的一项
   * @method removeItem
   * @param {string|Array} key 仓库中的键
   * @param {boolean} [isTemp=true] 是否删除临时的仓库，默认为 true
   *
   * @memberOf VuexStorage
   */
  removeItem(key, isTemp = true) {
    const { isPass } = this.checkItem(key, isTemp)

    if (isPass) {
      VuexStorage.loopCallBack(store => {
        this.$store.unregisterModule(store.concat(key))
      }, isTemp)
    }
  }

  /**
   * 检查项是否存在
   * @method checkItem
   * @param {string|Array} key 检查的项的键值
   * @param {boolean} [isTemp=true] 是否检查临时仓库
   * @param {boolean} [isWrite=false] 当项不存在是否自动创建
   * @param {Object} [updateItem=undefined] 更新项
   * @returns {boolean|Object} 返回
   *
   */
  checkItem(key, isTemp = true, isWrite = false, updateItem = undefined) {
    const keys = VuexStorage.getStorageRelation()[this.getBackUp ? 1 : 0].concat(key)
    const itemRelation = []

    let parentItem = this.$store.state

    const isPass = !(keys.some(k => {
      let item = parentItem[k]

      if (!item) {
        itemRelation.push(k)
        if (!isWrite) {
          return true
        }
        this.innerOperation = true
        this.setItem(itemRelation, {}, isTemp)
        item = parentItem[k]
      }

      parentItem = item
      return false
    }))

    if (!isPass) {
      parentItem = undefined
    } else if (updateItem) {
      const { lastKey, value } = updateItem
      parentItem[lastKey] = value
    }

    return {
      isPass,
      item: parentItem,
    }
  }

  /**
   * 更新指定项
   * @method updateItem
   * @param {string|Array} key 仓库中的键
   * @param {any} value 需要修改的值
   * @param {boolean} [isTemp=true] 是否存储到临时的仓库，默认为 true
   *
   * @memberOf VuexStorage
   */
  updateItem(key, value, isTemp = true) {
    this.$store.commit({
      type: this.UPDATE_ITEM,
      key,
      value,
      isTemp
    })
  }

  /**
   * 清除所有
   * @method clear
   * @param {boolean} [isTemp=true] 是否清楚临时的仓库，默认为 true
   *
   * @memberOf VuexStorage
   */
  clear(isTemp = true) {
    this.removeItem([], isTemp)
    this.innerOperation = true
    this.setItem([], {}, isTemp)
  }


  /**
   * 初始化操作
   * @method init
   * @param {boolean} [isTemp=true] 是否清楚临时的仓库，默认为 true
   *
   * @memberOf VuexStorage
   */
  initStorage() {
    this.clear()
    this.clear(false)
  }

  /**
   * 循环调用回调函数
   * @static
   * @method loopCallBack
   * @param {Function} cb 每个仓库的回调使用方法
   * @param {boolean} isTemp 是否使用临时仓库
   *
   * @memberOf VuexStorage
   */
  static loopCallBack(cb, isTemp) {
    const stores = VuexStorage.getStorageRelation(isTemp)
    stores.forEach((store, index) => {
      cb(store, index)
    })
  }

  /**
   * 获取仓库基础关系
   * @static
   * @method getStorageRelation
   * @return {Array} 仓库的基础关系
   * @memberOf VuexStorage
   */
  static getStorageRelation() {
    return [
      ['using'],
      ['backup']
    ]
  }

}

export const createStore = (
  Vue,
  router,
  routeWhiteList = [],
  cb = () => Promise.resolve,
) => {
  Vue.use(Vuex)
  // Vue.use(require('./storage-http'))
  const storage = new VuexStorage(routeWhiteList)

  const store = new Vuex.Store(storage.generateVuexInstance())

  storage.$store = store

  storage.initStorage()

  Vue.prototype.storage = storage

  return store
}
